from dydx3.eth_signing.api_key_action import SignApiKeyAction
from dydx3.eth_signing.onboarding_action import SignOnboardingAction
from dydx3.eth_signing.signers import SignWithKey
from dydx3.eth_signing.signers import SignWithWeb3



from datetime import datetime
import json
import random

import dateutil.parser as dp


def generate_query_path(url, params):
    entries = params.items()
    if not entries:
        return url

    paramsString = '&'.join('{key}={value}'.format(
        key=x[0], value=x[1]) for x in entries if x[1] is not None)
    if paramsString:
        return url + '?' + paramsString

    return url


def json_stringify(data):
    return json.dumps(data, separators=(',', ':'))


def random_client_id():
    # TODO: Backend should treat client ID as a string, not a number.
    # NOTE: For now, need to make sure to remove leading zeros.
    return str(int(float(str(random.random())[2:])))


def generate_now_iso():
    return datetime.utcnow().strftime(
        '%Y-%m-%dT%H:%M:%S.%f',
    )[:-3] + 'Z'


def iso_to_epoch_seconds(iso):
    return dp.parse(iso).timestamp()


def epoch_seconds_to_iso(epoch):
    return datetime.utcfromtimestamp(epoch).strftime(
        '%Y-%m-%dT%H:%M:%S.%f',
    )[:-3] + 'Z'


def remove_nones(original):
    return {k: v for k, v in original.items() if v is not None}



from web3 import Web3

from dydx3.constants import NETWORK_ID_MAINNET
from dydx3.eth_signing import SignWithWeb3
from dydx3.eth_signing import SignWithKey
from dydx3.modules.api_keys import ApiKeys
from dydx3.modules.eth import Eth
from dydx3.modules.private import Private
from dydx3.modules.public import Public
from dydx3.modules.onboarding import Onboarding
from dydx3.starkex.helpers import private_key_to_public_key_pair_hex
from dydx3.starkex.starkex_resources.cpp_signature import (
    get_cpp_lib,
)


class Client(object):

    def __init__(
        self,
        host,
        api_timeout=3000,  # TODO: Actually use this.
        default_ethereum_address=None,
        eth_private_key=None,
        eth_send_options=None,
        network_id=None,
        stark_private_key=None,
        stark_public_key=None,
        stark_public_key_y_coordinate=None,
        web3=None,
        web3_account=None,
        web3_provider=None,
        api_key_credentials=None,
        crypto_c_exports_path=None,
    ):
        # Remove trailing '/' if present, from host.
        if host.endswith('/'):
            host = host[:-1]

        self.host = host
        self.api_timeout = api_timeout
        self.eth_send_options = eth_send_options or {}
        self.stark_private_key = stark_private_key
        self.api_key_credentials = api_key_credentials
        self.stark_public_key_y_coordinate = stark_public_key_y_coordinate

        self.web3 = None
        self.eth_signer = None
        self.default_address = None
        self.network_id = None

        if crypto_c_exports_path is not None:
            get_cpp_lib(crypto_c_exports_path)

        if web3 is not None or web3_provider is not None:
            if isinstance(web3_provider, str):
                web3_provider = Web3.HTTPProvider(web3_provider)
            self.web3 = web3 or Web3(web3_provider)
            self.eth_signer = SignWithWeb3(self.web3)
            self.default_address = self.web3.eth.defaultAccount or None
            self.network_id = self.web3.net.version

        if eth_private_key is not None or web3_account is not None:
            # May override web3 or web3_provider configuration.
            key = eth_private_key or web3_account.key
            self.eth_signer = SignWithKey(key)
            self.default_address = self.eth_signer.address

        self.default_address = default_ethereum_address or self.default_address
        self.network_id = int(
            network_id or self.network_id or NETWORK_ID_MAINNET
        )

        # Initialize the public module. Other modules are initialized on
        # demand, if the necessary configuration options were provided.
        self._public = Public(host)
        self._private = None
        self._api_keys = None
        self._eth = None
        self._onboarding = None

        # Derive the public keys.
        if stark_private_key is not None:
            self.stark_public_key, self.stark_public_key_y_coordinate = (
                private_key_to_public_key_pair_hex(stark_private_key)
            )
            if (
                stark_public_key is not None and
                stark_public_key != self.stark_public_key
            ):
                raise ValueError('STARK public/private key mismatch')
            if (
                stark_public_key_y_coordinate is not None and
                stark_public_key_y_coordinate !=
                    self.stark_public_key_y_coordinate
            ):
                raise ValueError('STARK public/private key mismatch (y)')
        else:
            self.stark_public_key = stark_public_key
            self.stark_public_key_y_coordinate = stark_public_key_y_coordinate

        # Generate default API key credentials if needed and possible.
        if (
            self.eth_signer and
            self.default_address and
            not self.api_key_credentials
        ):
            # This may involve a web3 call, so recover on failure.
            try:
                self.api_key_credentials = (
                    self.onboarding.recover_default_api_key_credentials(
                        ethereum_address=self.default_address,
                    )
                )
            except Exception as e:
                print(
                    'Warning: Failed to derive default API key credentials:',
                    e,
                )

    @property
    def public(self):
        '''
        Get the public module, used for interacting with public endpoints.
        '''
        return self._public

    @property
    def private(self):
        '''
        Get the private module, used for interacting with endpoints that
        require API-key auth.
        '''
        if not self._private:
            if self.api_key_credentials:
                self._private = Private(
                    host=self.host,
                    network_id=self.network_id,
                    stark_private_key=self.stark_private_key,
                    default_address=self.default_address,
                    api_key_credentials=self.api_key_credentials,
                )
            else:
                raise Exception(
                    'Private endpoints not supported ' +
                    'since api_key_credentials were not specified',
                )
        return self._private

    @property
    def api_keys(self):
        '''
        Get the api_keys module, used for managing API keys. Requires
        Ethereum key auth.
        '''
        if not self._api_keys:
            if self.eth_signer:
                self._api_keys = ApiKeys(
                    host=self.host,
                    eth_signer=self.eth_signer,
                    network_id=self.network_id,
                    default_address=self.default_address,
                )
            else:
                raise Exception(
                    'API keys module is not supported since no Ethereum ' +
                    'signing method (web3, web3_account, web3_provider) was ' +
                    'provided',
                )
        return self._api_keys

    @property
    def onboarding(self):
        '''
        Get the onboarding module, used to create a new user. Requires
        Ethereum key auth.
        '''
        if not self._onboarding:
            if self.eth_signer:
                self._onboarding = Onboarding(
                    host=self.host,
                    eth_signer=self.eth_signer,
                    network_id=self.network_id,
                    default_address=self.default_address,
                    stark_public_key=self.stark_public_key,
                    stark_public_key_y_coordinate=(
                        self.stark_public_key_y_coordinate
                    ),
                )
            else:
                raise Exception(
                    'Onboarding is not supported since no Ethereum ' +
                    'signing method (web3, web3_account, web3_provider) was ' +
                    'provided',
                )
        return self._onboarding

    @property
    def eth(self):
        '''
        Get the eth module, used for interacting with Ethereum smart contracts.
        '''
        if not self._eth:
            eth_private_key = getattr(self.eth_signer, '_private_key', None)
            if self.web3 and eth_private_key:
                self._eth = Eth(
                    web3=self.web3,
                    network_id=self.network_id,
                    eth_private_key=eth_private_key,
                    default_address=self.default_address,
                    stark_public_key=self.stark_public_key,
                    send_options=self.eth_send_options,
                )
            else:
                raise Exception(
                    'Eth module is not supported since neither web3 ' +
                    'nor web3_provider was provided OR since neither ' +
                    'eth_private_key nor web3_account was provided',
                )
        return self._eth



import json

import requests

from dydx3.errors import DydxApiError
from dydx3.helpers.request_helpers import remove_nones

# TODO: Use a separate session per client instance.
session = requests.session()
session.headers.update({
    'Accept': 'application/json',
    'Content-Type': 'application/json',
    'User-Agent': 'dydx/python',
})


def request(uri, method, headers=None, data_values={}):
    response = send_request(
        uri,
        method,
        headers,
        data=json.dumps(
            remove_nones(data_values)
        )
    )
    if not str(response.status_code).startswith('2'):
        raise DydxApiError(response)
    return response.json() if response.content else '{}'


def send_request(uri, method, headers=None, **kwargs):
    return getattr(session, method)(uri, headers=headers, **kwargs)



from web3 import Web3

from dydx3 import Client
from dydx3.constants import NETWORK_ID_MAINNET
from dydx3.constants import NETWORK_ID_ROPSTEN

from tests.constants import DEFAULT_HOST

GANACHE_PRIVATE_KEY = (
    '0x4f3edf983ac636a65a842ce7c78d9aa706d3b113bce9c46f30d7d21715b23b1d'
)

EXPECTED_API_KEY_CREDENTIALS_MAINNET = {
    'key': '50fdcaa0-62b8-e827-02e8-a9520d46cb9f',
    'secret': 'rdHdKDAOCa0B_Mq-Q9kh8Fz6rK3ocZNOhKB4QsR9',
    'passphrase': '12_1LuuJMZUxcj3kGBWc',
}
EXPECTED_STARK_PRIVATE_KEY_MAINNET = (
    '0x170d807cafe3d8b5758f3f698331d292bf5aeb71f6fd282f0831dee094ee891'
)
EXPECTED_API_KEY_CREDENTIALS_ROPSTEN = {
    'key': '9c1d91a5-0a30-1ed4-2d3d-b840a479b965',
    'secret': 'hHYEswFe5MHMm8gFb81Jas9b7iLQUicsVv5YBRMY',
    'passphrase': '9z5Ew7m2DLQd87Xlk7Hd',
}
EXPECTED_STARK_PRIVATE_KEY_ROPSTEN = (
    '0x50505654b282eb3debadddeddfa1bc76545a6837dcd59d7d41f6a282a4bbccc'
)


class TestOnboarding():

    def test_derive_stark_key_on_mainnet_from_web3(self):
        web3 = Web3()  # Connect to a local Ethereum node.
        client = Client(
            host=DEFAULT_HOST,
            network_id=NETWORK_ID_MAINNET,
            web3=web3,
        )
        signer_address = web3.eth.accounts[0]
        stark_private_key = client.onboarding.derive_stark_key(signer_address)
        assert stark_private_key == EXPECTED_STARK_PRIVATE_KEY_MAINNET

    def test_recover_default_api_key_credentials_on_mainnet_from_web3(self):
        web3 = Web3()  # Connect to a local Ethereum node.
        client = Client(
            host=DEFAULT_HOST,
            network_id=NETWORK_ID_MAINNET,
            web3=web3,
        )
        signer_address = web3.eth.accounts[0]
        api_key_credentials = (
            client.onboarding.recover_default_api_key_credentials(
                signer_address,
            )
        )
        assert api_key_credentials == EXPECTED_API_KEY_CREDENTIALS_MAINNET

    def test_derive_stark_key_on_ropsten_from_web3(self):
        web3 = Web3()  # Connect to a local Ethereum node.
        client = Client(
            host=DEFAULT_HOST,
            network_id=NETWORK_ID_ROPSTEN,
            web3=web3,
        )
        signer_address = web3.eth.accounts[0]
        stark_private_key = client.onboarding.derive_stark_key(signer_address)
        assert stark_private_key == EXPECTED_STARK_PRIVATE_KEY_ROPSTEN

    def test_recover_default_api_key_credentials_on_ropsten_from_web3(self):
        web3 = Web3()  # Connect to a local Ethereum node.
        client = Client(
            host=DEFAULT_HOST,
            network_id=NETWORK_ID_ROPSTEN,
            web3=web3,
        )
        signer_address = web3.eth.accounts[0]
        api_key_credentials = (
            client.onboarding.recover_default_api_key_credentials(
                signer_address,
            )
        )
        assert api_key_credentials == EXPECTED_API_KEY_CREDENTIALS_ROPSTEN

    def test_derive_stark_key_on_mainnet_from_priv(self):
        client = Client(
            host=DEFAULT_HOST,
            network_id=NETWORK_ID_MAINNET,
            eth_private_key=GANACHE_PRIVATE_KEY,
            api_key_credentials={'key': 'value'},
        )
        signer_address = client.default_address
        stark_private_key = client.onboarding.derive_stark_key(signer_address)
        assert stark_private_key == EXPECTED_STARK_PRIVATE_KEY_MAINNET

    def test_recover_default_api_key_credentials_on_mainnet_from_priv(self):
        client = Client(
            host=DEFAULT_HOST,
            network_id=NETWORK_ID_MAINNET,
            eth_private_key=GANACHE_PRIVATE_KEY,
        )
        signer_address = client.default_address
        api_key_credentials = (
            client.onboarding.recover_default_api_key_credentials(
                signer_address,
            )
        )
        assert api_key_credentials == EXPECTED_API_KEY_CREDENTIALS_MAINNET

    def test_derive_stark_key_on_ropsten_from_priv(self):
        client = Client(
            host=DEFAULT_HOST,
            network_id=NETWORK_ID_ROPSTEN,
            eth_private_key=GANACHE_PRIVATE_KEY,
        )
        signer_address = client.default_address
        stark_private_key = client.onboarding.derive_stark_key(signer_address)
        assert stark_private_key == EXPECTED_STARK_PRIVATE_KEY_ROPSTEN

    def test_recover_default_api_key_credentials_on_ropsten_from_priv(self):
        client = Client(
            host=DEFAULT_HOST,
            network_id=NETWORK_ID_ROPSTEN,
            eth_private_key=GANACHE_PRIVATE_KEY,
        )
        signer_address = client.default_address
        api_key_credentials = (
            client.onboarding.recover_default_api_key_credentials(
                signer_address,
            )
        )
        assert api_key_credentials == EXPECTED_API_KEY_CREDENTIALS_ROPSTEN



from dydx3.starkex.starkex_resources.cpp_signature import (
  check_cpp_lib_path, cpp_hash, cpp_sign, cpp_verify
)
from dydx3.starkex.starkex_resources.python_signature import (
  py_sign, ECSignature, ECPoint, py_verify, py_pedersen_hash
)
from typing import Optional, Union

def sign(msg_hash: int, priv_key: int, seed: Optional[int] = None) -> ECSignature:
  if check_cpp_lib_path():
    return cpp_sign(msg_hash=msg_hash, priv_key=priv_key, seed=seed)

  return py_sign(msg_hash=msg_hash, priv_key=priv_key, seed=seed)

def verify(msg_hash: int, r: int, s: int, public_key: Union[int, ECPoint]) -> bool:
  if check_cpp_lib_path():
    return cpp_verify(msg_hash=msg_hash, r=r, s=s, stark_key=public_key)

  return py_verify(msg_hash=msg_hash, r=r, s=s, public_key=public_key)

def get_hash(*elements: int) -> int:
  if check_cpp_lib_path():
    return cpp_hash(*elements)

  return py_pedersen_hash(*elements)



from dydx3.constants import MARKET_ETH_USD
from dydx3.constants import NETWORK_ID_ROPSTEN
from dydx3.constants import ORDER_SIDE_BUY
from dydx3.helpers.request_helpers import iso_to_epoch_seconds
from dydx3.starkex.order import SignableOrder

# Test data where the public key y-coordinate is odd.
MOCK_PUBLIC_KEY = (
    '3b865a18323b8d147a12c556bfb1d502516c325b1477a23ba6c77af31f020fd'
)
MOCK_PRIVATE_KEY = (
    '58c7d5a90b1776bde86ebac077e053ed85b0f7164f53b080304a531947f46e3'
)
MOCK_SIGNATURE = (
    '00cecbe513ecdbf782cd02b2a5efb03e58d5f63d15f2b840e9bc0029af04e8dd' +
    '0090b822b16f50b2120e4ea9852b340f7936ff6069d02acca02f2ed03029ace5'
)

# Test data where the public key y-coordinate is even.
MOCK_PUBLIC_KEY_EVEN_Y = (
    '5c749cd4c44bdc730bc90af9bfbdede9deb2c1c96c05806ce1bc1cb4fed64f7'
)
MOCK_SIGNATURE_EVEN_Y = (
    '00fc0756522d78bef51f70e3981dc4d1e82273f59cdac6bc31c5776baabae6ec' +
    '0158963bfd45d88a99fb2d6d72c9bbcf90b24c3c0ef2394ad8d05f9d3983443a'
)

# Mock order params.
ORDER_PARAMS = {
    "network_id": NETWORK_ID_ROPSTEN,
    "market": MARKET_ETH_USD,
    "side": ORDER_SIDE_BUY,
    "position_id": 12345,
    "human_size": '145.0005',
    "human_price": '350.00067',
    "limit_fee": '0.125',
    "client_id": (
        'This is an ID that the client came up with ' +
        'to describe this order'
    ),
    "expiration_epoch_seconds": iso_to_epoch_seconds(
        '2020-09-17T04:15:55.028Z',
    ),
}


class TestOrder():

    def test_sign_order(self):
        order = SignableOrder(**ORDER_PARAMS)
        signature = order.sign(MOCK_PRIVATE_KEY)
        assert signature == MOCK_SIGNATURE

    def test_verify_signature_odd_y(self):
        order = SignableOrder(**ORDER_PARAMS)
        assert order.verify_signature(MOCK_SIGNATURE, MOCK_PUBLIC_KEY)

    def test_verify_signature_even_y(self):
        order = SignableOrder(**ORDER_PARAMS)
        assert order.verify_signature(
            MOCK_SIGNATURE_EVEN_Y,
            MOCK_PUBLIC_KEY_EVEN_Y,
        )

    def test_starkware_representation(self):
        order = SignableOrder(**ORDER_PARAMS)
        starkware_order = order.to_starkware()
        assert starkware_order.quantums_amount_synthetic == 145000500000
        assert starkware_order.quantums_amount_collateral == 50750272151
        assert starkware_order.quantums_amount_fee == 6343784019

        # Order expiration should be rounded up and should have a buffer added.
        assert starkware_order.expiration_epoch_hours == 444701

    def test_convert_order_fee_edge_case(self):
        order = SignableOrder(
            **dict(
                ORDER_PARAMS,
                limit_fee='0.000001999999999999999999999999999999999999999999',
            ),
        )
        starkware_order = order.to_starkware()
        assert starkware_order.quantums_amount_fee == 50751

    def test_order_expiration_boundary_case(self):
        order = SignableOrder(
            **dict(
                ORDER_PARAMS,
                expiration_epoch_seconds=iso_to_epoch_seconds(
                    # Less than one second after the start of the hour.
                    '2021-02-24T16:00:00.407Z',
                ),
            ),
        )
        assert order.to_starkware().expiration_epoch_hours == 448553



from dydx3.starkex.helpers import fact_to_condition
from dydx3.starkex.helpers import generate_private_key_hex_unsafe
from dydx3.starkex.helpers import get_transfer_erc20_fact
from dydx3.starkex.helpers import nonce_from_client_id
from dydx3.starkex.helpers import private_key_from_bytes
from dydx3.starkex.helpers import private_key_to_public_hex
from dydx3.starkex.helpers import private_key_to_public_key_pair_hex


class TestHelpers():

    def test_nonce_from_client_id(self):
        assert nonce_from_client_id('') == 2018687061
        assert nonce_from_client_id('1') == 3079101259
        assert nonce_from_client_id('a') == 2951628987
        assert nonce_from_client_id(
            'A really long client ID used to identify an order or withdrawal',
        ) == 2913863714
        assert nonce_from_client_id(
            'A really long client ID used to identify an order or withdrawal!',
        ) == 230317226

    def test_get_transfer_erc20_fact(self):
        assert get_transfer_erc20_fact(
            recipient='0x1234567890123456789012345678901234567890',
            token_decimals=3,
            human_amount=123.456,
            token_address='0xaAaAaAaaAaAaAaaAaAAAAAAAAaaaAaAaAaaAaaAa',
            salt=int('0x1234567890abcdef', 16),
        ).hex() == (
            '34052387b5efb6132a42b244cff52a85a507ab319c414564d7a89207d4473672'
        )

    def test_fact_to_condition(self):
        fact = bytes.fromhex(
            'cf9492ae0554c642b57f5d9cabee36fb512dd6b6629bdc51e60efb3118b8c2d8'
        )
        condition = fact_to_condition(
            '0xe4a295420b58a4a7aa5c98920d6e8a0ef875b17a',
            fact,
        )
        assert hex(condition) == (
            '0x4d794792504b063843afdf759534f5ed510a3ca52e7baba2e999e02349dd24'
        )

    def test_generate_private_key_hex_unsafe(self):
        assert (
            generate_private_key_hex_unsafe() !=
            generate_private_key_hex_unsafe()
        )

    def test_private_key_from_bytes(self):
        assert (
            private_key_from_bytes(b'0') ==
            '0x2242959533856f2a03f3c7d9431e28ef4fe5cb2a15038c37f1d76d35dc508b'
        )
        assert (
            private_key_from_bytes(b'a') ==
            '0x1d61128b46faa109512e0e00fe9adf5ff52047ed61718eeeb7c0525dfcd2f8e'
        )
        assert (
            private_key_from_bytes(
                b'really long input data for key generation with the '
                b'keyPairFromData() function'
            ) ==
            '0x7c4946831bde597b73f1d5721af9c67731eafeb75c1b8e92ac457a61819a29'
        )

    def test_private_key_to_public_hex(self):
        assert private_key_to_public_hex(
            '0x2242959533856f2a03f3c7d9431e28ef4fe5cb2a15038c37f1d76d35dc508b',
        ) == (
            '0x69a33d37101d7089b606f92e4b41553c237a474ad9d6f62eeb6708415f98f4d'
        )

    def test_private_key_to_public_key_pair_hex(self):
        x, y = private_key_to_public_key_pair_hex(
            '0x2242959533856f2a03f3c7d9431e28ef4fe5cb2a15038c37f1d76d35dc508b',
        )
        assert x == (
            '0x69a33d37101d7089b606f92e4b41553c237a474ad9d6f62eeb6708415f98f4d'
        )
        assert y == (
            '0x717e78b98a53888aa7685b91137fa01b9336ce7d25f874dbfb8d752c6ac610d'
        )



import uuid

NAMESPACE = uuid.UUID('0f9da948-a6fb-4c45-9edc-4685c3f3317d')


def get_user_id(address):
    return str(uuid.uuid5(NAMESPACE, address))


def get_account_id(
    address,
    accountNumber=0,
):
    return str(
        uuid.uuid5(
            NAMESPACE,
            get_user_id(address.lower()) + str(accountNumber),
        ),
    )



from web3 import Web3

from dydx3.constants import NETWORK_ID_MAINNET
from dydx3.eth_signing import util
from dydx3.eth_signing.sign_off_chain_action import SignOffChainAction

# On mainnet, include an extra onlySignOn parameter.
EIP712_ONBOARDING_ACTION_STRUCT = [
    {'type': 'string', 'name': 'action'},
    {'type': 'string', 'name': 'onlySignOn'},
]
EIP712_ONBOARDING_ACTION_STRUCT_STRING = (
    'dYdX(' +
    'string action,' +
    'string onlySignOn' +
    ')'
)
EIP712_ONBOARDING_ACTION_STRUCT_TESTNET = [
    {'type': 'string', 'name': 'action'},
]
EIP712_ONBOARDING_ACTION_STRUCT_STRING_TESTNET = (
    'dYdX(' +
    'string action' +
    ')'
)
EIP712_STRUCT_NAME = 'dYdX'

ONLY_SIGN_ON_DOMAIN_MAINNET = 'https://trade.dydx.exchange'


class SignOnboardingAction(SignOffChainAction):

    def get_eip712_struct(self):
        # On mainnet, include an extra onlySignOn parameter.
        if self.network_id == NETWORK_ID_MAINNET:
            return EIP712_ONBOARDING_ACTION_STRUCT
        else:
            return EIP712_ONBOARDING_ACTION_STRUCT_TESTNET

    def get_eip712_struct_name(self):
        return EIP712_STRUCT_NAME

    def get_eip712_message(
        self,
        **message,
    ):
        eip712_message = super(SignOnboardingAction, self).get_eip712_message(
            **message,
        )

        # On mainnet, include an extra onlySignOn parameter.
        if self.network_id == NETWORK_ID_MAINNET:
            eip712_message['message']['onlySignOn'] = (
                'https://trade.dydx.exchange'
            )

        return eip712_message

    def get_hash(
        self,
        action,
    ):
        # On mainnet, include an extra onlySignOn parameter.
        if self.network_id == NETWORK_ID_MAINNET:
            eip712_struct_str = EIP712_ONBOARDING_ACTION_STRUCT_STRING
        else:
            eip712_struct_str = EIP712_ONBOARDING_ACTION_STRUCT_STRING_TESTNET

        data = [
            [
                'bytes32',
                'bytes32',
            ],
            [
                util.hash_string(eip712_struct_str),
                util.hash_string(action),
            ],
        ]

        # On mainnet, include an extra onlySignOn parameter.
        if self.network_id == NETWORK_ID_MAINNET:
            data[0].append('bytes32')
            data[1].append(util.hash_string(ONLY_SIGN_ON_DOMAIN_MAINNET))

        struct_hash = Web3.solidityKeccak(*data)
        return self.get_eip712_hash(struct_hash)



import ctypes
import secrets
import os
from typing import Optional, Tuple
import json
import math
from dydx3.starkex.starkex_resources.python_signature import (
    inv_mod_curve_size,
)

PEDERSEN_HASH_POINT_FILENAME = os.path.join(
    os.path.dirname(__file__), 'pedersen_params.json')
PEDERSEN_PARAMS = json.load(open(PEDERSEN_HASH_POINT_FILENAME))

EC_ORDER = PEDERSEN_PARAMS['EC_ORDER']

FIELD_PRIME = PEDERSEN_PARAMS['FIELD_PRIME']

N_ELEMENT_BITS_ECDSA = math.floor(math.log(FIELD_PRIME, 2))
assert N_ELEMENT_BITS_ECDSA == 251


CPP_LIB_PATH = None
OUT_BUFFER_SIZE = 251

def get_cpp_lib(crypto_c_exports_path):
    global CPP_LIB_PATH
    CPP_LIB_PATH = ctypes.cdll.LoadLibrary(os.path.abspath(crypto_c_exports_path))
    # Configure argument and return types.
    CPP_LIB_PATH.Hash.argtypes = [
        ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p]
    CPP_LIB_PATH.Verify.argtypes = [
        ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p]
    CPP_LIB_PATH.Verify.restype = bool
    CPP_LIB_PATH.Sign.argtypes = [
        ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p]

def check_cpp_lib_path() -> bool:
  return CPP_LIB_PATH is not None

#########
# ECDSA #
#########

# A type for the digital signature.
ECSignature = Tuple[int, int]

#################
# CPP WRAPPERS #
#################

def cpp_hash(left, right) -> int:
    res = ctypes.create_string_buffer(OUT_BUFFER_SIZE)
    if CPP_LIB_PATH.Hash(
            left.to_bytes(32, 'little', signed=False),
            right.to_bytes(32, 'little', signed=False),
            res) != 0:
        raise ValueError(res.raw.rstrip(b'\00'))
    return int.from_bytes(res.raw[:32], 'little', signed=False)


def cpp_sign(msg_hash, priv_key, seed: Optional[int] = 32) -> ECSignature:
    """
    Note that this uses the secrets module to generate cryptographically strong random numbers.
    Note that the same seed will give a different signature compared with the sign function in
    signature.py.
    """
    res = ctypes.create_string_buffer(OUT_BUFFER_SIZE)
    random_bytes = secrets.token_bytes(seed)
    if CPP_LIB_PATH.Sign(
            priv_key.to_bytes(32, 'little', signed=False),
            msg_hash.to_bytes(32, 'little', signed=False),
            random_bytes, res) != 0:
        raise ValueError(res.raw.rstrip(b'\00'))
    w = int.from_bytes(res.raw[32:64], 'little', signed=False)
    s = inv_mod_curve_size(w)
    return (int.from_bytes(res.raw[:32], 'little', signed=False), s)


def cpp_verify(msg_hash, r, s, stark_key) -> bool:
    w =inv_mod_curve_size(s)
    assert 1 <= stark_key < 2**N_ELEMENT_BITS_ECDSA, 'stark_key = %s' % stark_key
    assert 1 <= msg_hash < 2**N_ELEMENT_BITS_ECDSA, 'msg_hash = %s' % msg_hash
    assert 1 <= r < 2**N_ELEMENT_BITS_ECDSA, 'r = %s' % r
    assert 1 <= w < EC_ORDER, 'w = %s' % w
    return CPP_LIB_PATH.Verify(
        stark_key.to_bytes(32, 'little', signed=False),
        msg_hash.to_bytes(32, 'little', signed=False),
        r.to_bytes(32, 'little', signed=False),
        w.to_bytes(32, 'little', signed=False))



from web3 import Web3

from dydx3.constants import NETWORK_ID_MAINNET
from dydx3.constants import OFF_CHAIN_ONBOARDING_ACTION
from dydx3.eth_signing import SignWithWeb3
from dydx3.eth_signing import SignWithKey
from dydx3.eth_signing import SignOnboardingAction

GANACHE_ADDRESS = '0x90F8bf6A479f320ead074411a4B0e7944Ea8c9C1'
GANACHE_PRIVATE_KEY = (
    '0x4f3edf983ac636a65a842ce7c78d9aa706d3b113bce9c46f30d7d21715b23b1d'
)

EXPECTED_SIGNATURE = (
    '0x0a30eea502e9805b95bd432fa1952e345dda3e9f72f7732aa00775865352e2b549'
    '29803c221e9e63861e4604fbc796a4e1a6ca23d49452338a3d7602aaf6d1841c00'
)


class TestOnboardingAction():

    def test_sign_via_local_node(self):
        web3 = Web3()  # Connect to a local Ethereum node.
        signer = SignWithWeb3(web3)

        action_signer = SignOnboardingAction(signer, NETWORK_ID_MAINNET)
        signature = action_signer.sign(
            GANACHE_ADDRESS,
            action=OFF_CHAIN_ONBOARDING_ACTION,
        )
        assert action_signer.verify(
            signature,
            GANACHE_ADDRESS,
            action=OFF_CHAIN_ONBOARDING_ACTION,
        )
        assert signature == EXPECTED_SIGNATURE

    def test_sign_via_account(self):
        web3 = Web3(None)
        web3_account = web3.eth.account.create()
        signer = SignWithKey(web3_account.key)

        action_signer = SignOnboardingAction(signer, NETWORK_ID_MAINNET)
        signature = action_signer.sign(
            signer.address,
            action=OFF_CHAIN_ONBOARDING_ACTION,
        )
        assert action_signer.verify(
            signature,
            signer.address,
            action=OFF_CHAIN_ONBOARDING_ACTION,
        )

    def test_sign_via_private_key(self):
        signer = SignWithKey(GANACHE_PRIVATE_KEY)

        action_signer = SignOnboardingAction(signer, NETWORK_ID_MAINNET)
        signature = action_signer.sign(
            GANACHE_ADDRESS,
            action=OFF_CHAIN_ONBOARDING_ACTION,
        )
        assert action_signer.verify(
            signature,
            GANACHE_ADDRESS,
            action=OFF_CHAIN_ONBOARDING_ACTION,
        )
        assert signature == EXPECTED_SIGNATURE



from dydx3.constants import NETWORK_ID_ROPSTEN
from dydx3.helpers.request_helpers import iso_to_epoch_seconds
from dydx3.starkex.withdrawal import SignableWithdrawal

MOCK_PUBLIC_KEY = (
    '3b865a18323b8d147a12c556bfb1d502516c325b1477a23ba6c77af31f020fd'
)
MOCK_PRIVATE_KEY = (
    '58c7d5a90b1776bde86ebac077e053ed85b0f7164f53b080304a531947f46e3'
)
MOCK_SIGNATURE = (
    '05e48c33f8205a5359c95f1bd7385c1c1f587e338a514298c07634c0b6c952ba' +
    '0687d6980502a5d7fa84ef6fdc00104db22c43c7fb83e88ca84f19faa9ee3de1'
)

# Mock withdrawal params.
WITHDRAWAL_PARAMS = {
    "network_id": NETWORK_ID_ROPSTEN,
    "position_id": 12345,
    "human_amount": '49.478023',
    "client_id": (
        'This is an ID that the client came up with ' +
        'to describe this withdrawal'
    ),
    "expiration_epoch_seconds": iso_to_epoch_seconds(
        '2020-09-17T04:15:55.028Z',
    ),
}


class TestWithdrawal():

    def test_sign_withdrawal(self):
        withdrawal = SignableWithdrawal(**WITHDRAWAL_PARAMS)
        signature = withdrawal.sign(MOCK_PRIVATE_KEY)
        assert signature == MOCK_SIGNATURE

    def test_verify_signature(self):
        withdrawal = SignableWithdrawal(**WITHDRAWAL_PARAMS)
        assert withdrawal.verify_signature(MOCK_SIGNATURE, MOCK_PUBLIC_KEY)

    def test_starkware_representation(self):
        withdrawal = SignableWithdrawal(**WITHDRAWAL_PARAMS)
        starkware_withdrawal = withdrawal.to_starkware()
        assert starkware_withdrawal.quantums_amount == 49478023

        # Order expiration should be rounded up and should have a buffer added.
        assert starkware_withdrawal.expiration_epoch_hours == 444533
